# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module to drive a recommender engine."""

import logging

from learner.data_worker.data_processor import DataProcessor, handle_label_encoding, filter_data
from learner.model_manager.similarities import Similarities
from learner.model_manager.classifiers import ClassifierHandler
from learner.data_worker.data_manager import TestDataManager, ValidationDataManager
from learner.model_manager.model_initializer import ModelLoader, get_pickle_dir
from learner.model_manager.prediction_manager import RecommenderSegmentPredictor
from learner.data_worker.output_handler import OutputHandler
from learner.data_worker.output_handler import pickle_object
from learner.engines.base_standard_engine import BaseStandard
from learner.utilities.timer import timeit
from learner.analysis.analysis import Analysis
from learner.utilities.templates import FEATURE_IMPORTANCE_PATH, FEATURE_IMPORTANCE_PATH_SEG, LOAD_MODEL_PATH_SEG


class BaseRecommender(BaseStandard):
    """The BaseRecommender class that implements the methods shared by all recommender engines."""

    def __init__(self, conf):
        super().__init__(conf)

    def fit_models(self, train_data, mdl, drop_cols):
        """Call appropriate methods to fit, cross-validate, or grid-search.

        :param train_data: the training dataset
        :param mdl: an item in the models_dict
        :param drop_cols: a list of columns that should be excluded from training
        :return: None
        """
        # save the train_features in case we need to use them later
        mdl["train_features"] = train_data.drop(drop_cols, axis=1).columns

        if self._conf.model.gs_activate:  # pragma: no cover
            mdl["model"] = ClassifierHandler.learner_grid_search(train_data.drop(drop_cols, axis=1),
                                                                 train_data[self._conf.column.target_col],
                                                                 mdl,
                                                                 self._conf.model.gs_kfold_params,
                                                                 self._conf.model.gs_options)
        elif self._conf.model.cv_activate:  # pragma: no cover
            mdl["model"] = ClassifierHandler.learner_cross_validation(train_data.drop(drop_cols, axis=1),
                                                                      train_data[self._conf.column.target_col],
                                                                      mdl,
                                                                      self._conf.model.cv_kfold_params,
                                                                      self._conf.model.cv_options)

        else:
            mdl["model"] = ClassifierHandler.learner_fit(train_data.drop(drop_cols, axis=1),
                                                         train_data[self._conf.column.target_col],
                                                         mdl)

    def get_mlb_sim_matrix(self, train_data):
        """Compute the items that are needed for training and making predictions.

        :return: the updated training data, mlb data, mlb object, and le object
        """
        # get the min support
        min_support = RecommenderHelper.get_min_support(train_data,
                                                        self._conf.column.target_col,
                                                        self._conf.recommender.support_cutoff,
                                                        self._conf.recommender.max_num_category)
        # apply the length and support constraints to filter the data for training
        train_data = RecommenderHelper.apply_length_support_constraint(train_data,
                                                                       self._conf.column.target_col,
                                                                       self._conf.recommender.max_length,
                                                                       self._conf.recommender.min_length,
                                                                       min_support)

        # perform a multilabel binarization of the target for computing similarity scores and extension
        # note that these calculations are done here because the target will be label encoded for training
        mlb_data, mlb_obj = DataProcessor.multilabel_binarizer(train_data,
                                                               self._conf.column.target_col,
                                                               self._conf.column.id_cols)

        # compute the similarity matrix
        sim_matrix = Similarities.compute_similarities(mlb_data[mlb_obj.classes_],
                                                       self._conf.similarities.metric,
                                                       return_dataframe=True,
                                                       cols=mlb_obj.classes_)

        train_data, le = handle_label_encoding(train_data, [self._conf.column.target_col])

        return train_data, mlb_data, mlb_obj, sim_matrix, le


class Recommender(BaseRecommender):
    def __init__(self, conf):
        """Take a conf object (from the Configuration class) to run a recommender engine without support for segmentation.

        :param conf: a conf object
        """
        super().__init__(conf)
        # set the cols that must be dropped from train data
        self._drop_cols = self._conf.column.drop_from_train + [self._conf.column.target_col]

    @timeit("build the models")
    def run_engine(self):
        """The main function for driving the recommender engine

        :return: None
        """
        logging.info("Building the recommendation models...")
        if self._conf.model.train_models:
            train_data, mlb_data, mlb_obj, sim_matrix, le = self.prepare_train_get_mlb_sim_matrix()
            self.train_models(train_data, mlb_data, mlb_obj, sim_matrix, le)
            self.predict()
        else:
            logging.info("Loading saved model...")
            self.load_models(self._conf.workspace.path)

        if self._conf.validation.validate and self._conf.data.test_prediction_activate:
            self.validate_output()

    @timeit("train the recommendation model(s)")
    def train_models(self, train_data, mlb_data, mlb_obj, sim_matrix, le):
        """Train a classifier for the recommender engine.

        :param train_data: the training dataset
        :param mlb_data: the data that is obtained from MultiLabelBinarizer
        :param mlb_obj: the MultiLabelBinarizer object
        :param sim_matrix: the similarity matrix obtained from similarity calculations
        :param le: the LabelEncoder object
        :return: None
        """
        logging.info("Building the models...")

        for tag, mdl in self.models_dict.items():
            logging.info("Fitting %s model", tag)

            _, mdl["mlb_obj"], mdl["sim_matrix"], mdl["le"] = mlb_data, mlb_obj, sim_matrix, le
            # train the model and save it into model dictionary
            # depending on the user input, either do grid_search, cross_validation, or a single fit
            self.fit_models(train_data, mdl, self._drop_cols)

            # save models
            if self._conf.model.save_models:
                file_name = self._conf.model.models_dict[tag]["path"] + self._conf.workspace.name + "_" + str(tag) \
                            + str(self._conf.sep_timetag) + ".sav"

                pickle_object(self.conf, self._conf.model.models_dict[tag]["path"], file_name, dict({tag: mdl}),
                              self.processor, self.feature_engineering)

            # handle feature importance if requested
            if self._conf.analysis.importance_activate:
                analysis = Analysis(self._conf)
                filename = FEATURE_IMPORTANCE_PATH.format(
                    output_name=self._conf.workspace.name,
                    tag=str(tag),
                    sep_timetag=str(self._conf.sep_timetag)
                )
                analysis.handle_feature_importance(mdl, filename)

    def prepare_train_get_mlb_sim_matrix(self):
        """Prepare the training dataset and compute the items that are needed for training and making predictions.

        :return: the updated training data, mlb data, mlb object, and le object
        """
        return self.get_mlb_sim_matrix(self.data)


class RecommenderSegment(BaseRecommender):
    def __init__(self, conf):
        """Take a conf object (from the Configuration class) to run a recommender engine with support for segmentation.

        :param conf: a conf object
        """
        super().__init__(conf)
        self._drop_cols = self._conf.column.drop_from_train + [self._conf.column.target_col] + \
            [self._conf.segmenter.test_name] + [self._conf.segmenter.test_col]

    @timeit("build the models")
    def run_engine(self):
        """The main function for driving the engine.

        :return: None
        """
        logging.info("Building the recommendation models...")
        if self._conf.model.train_models:
            for seg_id, border in enumerate(self._conf.segmenter.seg_list):
                train_data, mlb_data, mlb_obj, sim_matrix, le = self.prepare_train_get_mlb_sim_matrix(border)
                self.train_models(seg_id, border, train_data, mlb_data, mlb_obj, sim_matrix, le)

                self.predict(seg_id, border)
        else:
            logging.info("Loading saved model...")
            self.load_models(self._conf.workspace.path)

        # concatenate the output and write the results into a single file
        if self._conf.workspace.concat_output and self._conf.data.test_prediction_activate:
            output = OutputHandler(self._conf, data_type="test")
            output.concat_save_csv(self._conf.model.models_dict, self._conf.segmenter.seg_list,
                                   self._conf.workspace.name, self._conf.sep_timetag)

        if self._conf.validation.validate and self._conf.data.test_prediction_activate:
            self.validate_output()

    @timeit("train the recommendation model(s)")
    def train_models(self, seg_id, seg_border, train_data, mlb_data, mlb_obj, sim_matrix, le):
        """Train a classifier for the recommender engine.

        :param seg_id: the seg id
        :param seg_border: the upper boundary of the seg
        :param train_data: the training dataset
        :param mlb_data: the data that is obtained from MultiLabelBinarizer
        :param mlb_obj: the MultiLabelBinarizer object
        :param sim_matrix: the similarity matrix obtained from similarity calculations
        :param le: the LabelEncoder object
        :return: None
        """
        logging.info("**************************************************************************")
        logging.info("Training model on segment_id: %i, segment_border: %s", seg_id, seg_border)
        logging.info("There are %i rows of data in this segment", train_data.shape[0])

        for tag, mdls in self.models_dict.items():
            mdl = mdls.get(seg_id, mdls)
            logging.info("Fitting %s model", tag)

            _, mdl["mlb_obj"], mdl["sim_matrix"], mdl["le"] = mlb_data, mlb_obj, sim_matrix, le
            drop_cols = self._drop_cols + [self._conf.segmenter.train_name]
            # train the model and save it into model dictionary
            # depending on the user input, either do grid_search, cross_validation, or a single fit
            self.fit_models(train_data, mdl, drop_cols)

            # save models
            if self._conf.model.save_models:
                file_name = mdl["path"] + self._conf.workspace.name + "_" + str(tag) + "_" \
                            + str(seg_id) + str(self._conf.sep_timetag) + ".sav"

                pickle_object(self.conf, mdl["path"], file_name, dict({tag: mdl}),
                              self.processor, self.feature_engineering)

            # handle feature importance if requested
            if self._conf.analysis.importance_activate:
                analysis = Analysis(self._conf)
                filename = FEATURE_IMPORTANCE_PATH_SEG.format(
                    output_name=self._conf.workspace.name,
                    tag=str(tag),
                    seg_id=str(seg_id),
                    sep_timetag=str(self._conf.sep_sep_timetag)
                )
                analysis.handle_feature_importance(mdl, filename)

    @timeit("load the model(s)")
    def load_models(self, output_path):
        """Load pickled object from path_to_pickle if full directory is provided. Otherwise, it loads the latest pickled
         object.

        :param output_path: directory path to the pickled object
        :return: None
        """
        paths_to_pickle_dir, sample_pickle_files, prefixes = get_pickle_dir(output_path)
        loader = ModelLoader(self._conf)

        nrows_score = 0

        for path_to_pickle_dir, sample_pickle_file, prefix in zip(paths_to_pickle_dir, sample_pickle_files, prefixes):

            self._conf, self.models_dict, self.processor, self.feature_engineering, sep_timetag = loader.load_model(sample_pickle_file)

            for seg_id, seg_border in enumerate(self._conf.segmenter.seg_list):
                logging.info("**************************************************************************")
                logging.info("Making predictions for segment_id: %i, segment_border: %s", seg_id, seg_border)
                filename = LOAD_MODEL_PATH_SEG.format(
                    path=path_to_pickle_dir,
                    prefix=prefix,
                    seg_id=str(seg_id),
                    sep_timetag=str(sep_timetag),
                    ext=".sav"
                )
                self._conf, self.models_dict, self.processor, self.feature_engineering, sep_timetag = loader.load_model(filename)
                self.predict(seg_id, seg_border)

                # we need to make sure we add the number of predictions from all the segments
                nrows_score += self._conf.model.nrows_score

        # we need to assign the total number of predictions for later validation
        # we also don't want to double count if we have multiple models
        self._conf.model.nrows_score = nrows_score / len(paths_to_pickle_dir)

    @timeit("make predictions using the trained model(s)")
    def predict(self, seg_id=None, seg_border=None):
        """Call predict_validation if we need to score the models using validation data and call predict_test if we
        need to make predictions on test data.

        :param seg_id: segment id
        :param seg_border: the upper boundary of the segment
        :return: None
        """
        if self._conf.data.validation_score_activate:
            logging.info("Making predictions using validation data...")
            self.predict_validation(seg_id, seg_border)
        if self._conf.data.test_prediction_activate:
            logging.info("Making predictions using test data...")
            self.predict_test(seg_id, seg_border)

    def predict_validation(self, seg_id=None, seg_border=None):
        """Use the trained models to make prediction on validation data. Unlike test data, the entire validation data
        is loaded in memory when making predictions on validation data.

        :param seg_id: segment id
        :param seg_border: the upper boundary of the segment
        :return: None
        """
        validation_manager = ValidationDataManager(self._conf, self.processor, self.feature_engineering)
        validation_data = self.filter_data(validation_manager.data, self._conf.segmenter.test_name, seg_border)
        for tag, mdls in self.models_dict.items():
            mdl = mdls.get(seg_id, mdls)

            predictor = RecommenderSegmentPredictor(mdl, self._conf)
            prediction, data = predictor.make_predictions(validation_data, column_name=self._conf.data.validation_column_name)
            predictor.write_predictions(tag, prediction, data, mdl, seg_id=seg_id, data_type="validation")

    def predict_test(self, seg_id=None, seg_border=None):
        """Use the trained models to make predictions. Making predictions are performed by iterating through the
        test dataset. The test dataset can be very large and we process it in chunks.

        :param seg_id: the seg id
        :param seg_border: the upper boundary of the seg
        :return: None
        """
        test_manager = TestDataManager(self._conf, self.processor, self.feature_engineering)
        reader, indices = test_manager.get_reader_for_segment(self._conf.segmenter.test_col, seg_border)

        num_chunks = 0
        low_limit = 0
        for index, chunk in enumerate(reader):
            num_chunks += 1
            high_limit = low_limit + chunk.shape[0]
            select_rows = indices[(indices < high_limit) & (indices >= low_limit)]
            low_limit += chunk.shape[0]
            chunk = chunk.loc[select_rows, :]

            test_data = test_manager.get_reader_data(chunk)

            for tag, mdls in self.models_dict.items():
                mdl = mdls.get(seg_id, mdls)
                predictor = RecommenderSegmentPredictor(mdl, self._conf)
                prediction, data = predictor.make_predictions(test_data, column_name=self._conf.data.test_column_name)
                predictor.write_predictions(tag, prediction, data, index, mdl=mdl, seg_id=seg_id, data_type="test")

            # keep track of number of rows in test dataset for validation purposes
            self.conf.model.nrows_score += chunk.shape[0]

        output = OutputHandler(self._conf, data_type="test")
        output.concat_chunk_csv(num_chunks, self.models_dict, seg_id)

    def prepare_train_get_mlb_sim_matrix(self, seg_id):
        """Prepare the training dataset and compute the items that are needed for training and making predictions

        :param seg_id: the segment id
        :return: the updated training data, mlb data, mlb object, and le object
        """
        # get the train and test data for the segment. Note that the difference is in the segmenter base and
        # test names
        train_data = filter_data(self.data, self._conf.segmenter.train_name, seg_id, '==')

        return self.get_mlb_sim_matrix(train_data)


class RecommenderHelper:
    """Handle filtering the train data for recommender engine according to inputs provided by the user"""

    @staticmethod
    def get_min_support(data, col, cutoff=.20, max_num_baskets=100):
        """cutoff : the %-age cutoff in unique combinatons of service codes (sorted by decending relative counts)

        Returns the top [100*cutoff] percent of unique basket combinations.
        example:

        idx basket  count
        1   15,32   4321
        2   32      321
        3   12,32   320
        4   4,14,32 210
        5   15      184
        6   16,32   43
        7   37      32
        8   22,32   3
        9   14,32   2
        10  19      1

        get_min_basket_support(cutoff=.20) returns 321,(indended useage) meaning that only baskets with support >= 321 will be considered "valid"


        :param data: pandas dataframe (this is usually train data
        :param col: the column to perform the calculations on (this is usually target_col
        :param cutoff: the percentage of basket to keep for training. See the above explanation
        :param max_num_baskets: maximum number of basket to include
        :return: the minimum number of frequency a basket should appear to be included in train
        """
        unique_baskets = data.groupby(col)[col].count().sort_values(ascending=False)
        threshold_idx = int(len(unique_baskets) * cutoff)
        idx = min(max_num_baskets, threshold_idx)

        return unique_baskets[idx]

    @staticmethod
    def apply_length_support_constraint(data, col, max_length, min_length, min_support):
        """Filter a dataset according to the constraints set by user for recommender engine

        :param data: pandas dataframe
        :param col: the column to use for filtering. This is usually the target_col
        :param max_length: the maximum number of items that a basket can have
        :param min_length: the minimum number of items that a basket can have
        :param min_support: the minimum_support obtained from min_support method
        :return: the filtered data frame
        """

        data = data[(data[col].apply(len) <= max_length) & (data[col].apply(len) >= min_length)]
        sorted_support = data.groupby(col).size().sort_values(ascending=False).reset_index(name='count')
        sorted_support = sorted_support[col][sorted_support['count'] >= min_support]

        return data[data[col].isin(sorted_support)]



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module includes Learner classifiers. The classifiers inherit from the AbstractClassifiers. The instance methods
could wrap the sklearn methods or any other packages. The call to Learner classifiers are prefixed by learner\_.
This reduces the possibility of shadowing function calls to the external packages."""

import abc
import sys
import warnings
import logging
import numpy as np
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier, AdaBoostClassifier
from sklearn.ensemble import ExtraTreesClassifier, BaggingClassifier
from sklearn.tree import DecisionTreeClassifier, ExtraTreeClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.naive_bayes import GaussianNB, BernoulliNB, MultinomialNB
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis, QuadraticDiscriminantAnalysis
from xgboost import XGBClassifier
from lightgbm import LGBMClassifier
from sklearn.svm import SVC
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.neural_network import MLPClassifier
from catboost import CatBoostClassifier
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import cross_validate

from learner.utilities.progress_bar import GridSearchCVProgressBar


class AbstractClassifier:

    __metaclass__ = abc.ABCMeta

    def __init__(self):  # pragma: no cover
        self.mdl = None

    @staticmethod
    def learner_fit(X, y, mdl):  # pragma: no cover
        pass

    @staticmethod
    def learner_cross_validation(X, y, mdl, kfold_params, options):  # pragma: no cover
        pass

    @staticmethod
    def learner_grid_search(X, y, mdl, kfold_params, options):  # pragma: no cover
        pass

    @staticmethod
    def learner_predict(X, mdl, carry_data=True, full_data=None, cols_to_carry=None):  # pragma: no cover
        pass

    @staticmethod
    def learner_predict_proba(X, mdl, carry_data=True, full_data=None, cols_to_carry=None):  # pragma: no cover
        pass


class ClassifierHandler(AbstractClassifier):

    @abc.abstractmethod
    def __init__(self):
        super(AbstractClassifier, self).__init__()  # pragma: no cover

    @staticmethod
    def learner_fit(X, y, mdl):
        try:
            return getattr(sys.modules[__name__], mdl["type"]).learner_fit(X, y, mdl)
        except AttributeError:
            return StandardClassifiers.learner_fit(X, y, mdl)

    @staticmethod
    def learner_cross_validation(X, y, mdl, kfold_params, options):
        try:
            return getattr(sys.modules[__name__], mdl["type"]).learner_cross_validation(X, y, mdl, kfold_params, options)
        except AttributeError:
            return StandardClassifiers.learner_cross_validation(X, y, mdl, kfold_params, options)

    @staticmethod
    def learner_grid_search(X, y, mdl, kfold_params, options):
        try:
            return getattr(sys.modules[__name__], mdl["type"]).learner_grid_search(X, y, mdl, kfold_params, options)
        except AttributeError:
            return StandardClassifiers.learner_grid_search(X, y, mdl, kfold_params, options)

    @staticmethod
    def learner_predict(X, mdl, carry_data=True, full_data=None, cols_to_carry=None):
        try:
            return getattr(sys.modules[__name__], mdl["type"]).learner_predict(X, mdl, carry_data, full_data, cols_to_carry)
        except AttributeError:
            return StandardClassifiers.learner_predict(X, mdl, carry_data, full_data, cols_to_carry)

    @staticmethod
    def learner_predict_proba(X, mdl, carry_data=True, full_data=None, cols_to_carry=None):
        try:
            return getattr(sys.modules[__name__], mdl["type"]).learner_predict_proba(X, mdl, carry_data, full_data, cols_to_carry)
        except AttributeError:
            return StandardClassifiers.learner_predict_proba(X, mdl, carry_data, full_data, cols_to_carry)


class StandardClassifiers(AbstractClassifier):
    """A class for all standard classifiers. The standard classifiers are the ones that have "fit", "predict",
    and "predict_proba" methods and a few other features"""

    def __init__(self):
        super(AbstractClassifier, self).__init__()  # pragma: no cover

    @staticmethod
    def learner_fit(X, y, mdl):
        """Train a classifiers using the features (X) and the target y. Instantiation of the classifier object
            (sklearn's classifier) occurs here. The function return the classifier object

        :param X: the features for training
        :param y: the target column
        :param mdl: an item in the models_dicts
        :return: the classifiers object that contains the trained model
        """
        logging.info("Fitting a %s model...", mdl["type"])
        model = getattr(sys.modules[__name__], mdl["type"])(**mdl["params"])
        model.fit(X, y)
        # if the log level is debug, we log all the attributes
        logging.debug(vars(model))
        return model

    @staticmethod
    def learner_cross_validation(X, y, mdl, kfold_params, options):
        """Perform a kfold cross validation, report the average score and return the trained model.

        :param X: the features for training
        :param y: the target column
        :param mdl: an item in the models_dicts
        :param kfold_params: the parameters for kfold cross validation
        :param options: other options for doing a cross_validation
        :return: the classifiers object that contains the trained model
        """
        logging.info("Performing cross_validation for a %s model...", mdl["type"])

        model = getattr(sys.modules[__name__], mdl["type"])(**mdl["params"])
        skf = StratifiedKFold(**kfold_params)

        options["return_estimator"] = True
        cv_results = cross_validate(model, X, y=y, cv=skf, **options)
        if options.get("return_train_score", False):
            for score_type, score_values in cv_results.items():
                # the train scores keys start with train_. If we pass a list, it will contain the name of the metric.
                # If we pass nothing, it will be train_score
                if score_type.startswith("train_"):
                    avg_train_score = np.mean(score_values)
                    logging.info(f"The {score_type}s are {score_values} with average {avg_train_score}")

        avg_test_score = 0
        test_scores = 0
        # the test scores keys start with test_. If we pass a list, it will contain the name of the metric. If we
        # pass nothing, it will be test_score
        for score_type, score_values in cv_results.items():
            if score_type.startswith("test_"):
                test_scores = score_values
                avg_test_score = np.mean(cv_results[score_type])
                logging.info(f"The {score_type}s are {test_scores} with average {avg_test_score}")

        mdl["cv_score"] = avg_test_score
        return cv_results["estimator"][np.argmax(test_scores)]

    @staticmethod
    def learner_grid_search(X, y, mdl, kfold_params, options):
        """Perform an exhaustive grid search to find the best model.

        :param X: the features for training
        :param y: the target column
        :param mdl: an item in the Learner models_dict
        :param kfold_params: the parameters for kfold cross validation
        :param options: the input parameters for performing the grid_search
        :return: the best model
        """
        logging.info("Performing GridSearch for a %s model...", mdl["type"])
        model = getattr(sys.modules[__name__], mdl["type"])(**mdl["params"])
        skf = StratifiedKFold(**kfold_params)
        try:
            verbose = options["verbose"]
            if verbose > 0:  # pragma: no cover
                gs = GridSearchCV(model, mdl["params"], cv=skf, **options)
            else:  # pragma: no cover
                gs = GridSearchCVProgressBar(model, mdl["params"], cv=skf, **options)
        except KeyError:
            gs = GridSearchCVProgressBar(model, mdl["params"], cv=skf, **options)

        gs.fit(X, y)

        mdl["best_params"] = gs.best_params_
        logging.info("Best parameters are: %s", mdl["best_params"])
        logging.info("Best score is %f", gs.best_score_)
        return gs.best_estimator_

    @staticmethod
    def learner_predict(X, mdl, carry_data=True, full_data=None, cols_to_carry=None):
        """Make predictions using the trained model. In some situations, learner needs to save the predictions according to
        some identifiers (e.g. id_columns). In those situations, the functions also returns the data that the user has
        requested.

        :param X: the features to use for making predictions (obviously this should match with the features used for training)
        :param mdl: an item in the Learner models_dict
        :param carry_data: a flag to identify if the some selected data must be returned.
        :param full_data: the data set that contains the full data set. The selection is performed on this data set.
        :param cols_to_carry: the cols to select from the full data and return (this is usually the id_cols)
        :return: The prediction or the prediction along the selected columns

        Note: The reason for passing the full data is that the feature matrix usually does not contain id information.
        """
        logging.info("Making predictions using the %s model...", mdl["type"])
        if carry_data:
            if full_data is None:
                warnings.warn("carry_data was set true but no data was passed, will not return any data...", Warning)
            if cols_to_carry is None:
                warnings.warn("carry_data was set true but no cols_to_carry was passed, will not return any data...", Warning)
            if full_data is not None and cols_to_carry:
                return mdl["model"].predict(X), full_data[cols_to_carry]
        return mdl["model"].predict(X), None

    @staticmethod
    def learner_predict_proba(X, mdl, carry_data=True, full_data=None, cols_to_carry=None):
        """Predict the probabilities using the trained model. In some situations, learner needs to save the predictions
         according to some identifiers (e.g. id_columns). In those situations, the functions also return the data that
         the user has requested.

        :param X: the features to use for making predictions (obviously this should match with the features used for training)
        :param mdl: an item in the Learner models_dict
        :param carry_data: a flag to identify if the some selected data must be returned.
        :param full_data: the data set that contains the full data set. The selection is performed on this data set.
        :param cols_to_carry: the cols to select from the full data and return (this is usually the id_cols)
        :return: The prediction or the prediction along the selected columns

        Note: The reason for passing the full data is that the feature matrix usually does not contain id information.
        """
        logging.info("Making predictions using the %s model...", mdl["type"])
        if carry_data:
            if full_data is None:
                warnings.warn("carry_data was set true but no data was passed, will not return any data...", Warning)
            if cols_to_carry is None:
                warnings.warn("carry_data was set true but no cols_to_carry was passed, will not return any data...", Warning)
            if full_data is not None and cols_to_carry:
                if len(mdl["model"].classes_) == 2:
                    return mdl["model"].predict_proba(X)[:, 1], full_data[cols_to_carry]
                else:
                    return mdl["model"].predict_proba(X), full_data[cols_to_carry]
        if len(mdl["model"].classes_) == 2:
            return mdl["model"].predict_proba(X)[:, 1], None
        return mdl["model"].predict_proba(X), None



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import sys
import logging

from learner.data_worker.data_loader import get_value
from learner.configuration.supported_items import *
from learner.validator.input_validator import validate_intersection_cols, validate_subset_list
from learner.configuration.process import ProcessConfiguration
from learner.configuration.column import ColumnConfiguration
from learner.configuration.supported_items import SUPPORTED_DUPLICATES_FOR_SEGMENTERS


class SegmenterConfiguration:
    """Parse all input variables related to segmenter section. Learner currently supports static and
    dynamic segmentation."""

    def __init__(self, json_config, process: ProcessConfiguration, column: ColumnConfiguration):
        self._json_config = json_config
        self._process = process
        self._column = column
        self.activate = get_value(self._json_config, False, "segmenter", "activate")
        self.type = self.get_type()
        self.bin_size = get_value(self._json_config, 1, "segmenter", "params", "bin_size")
        self.max_num_bins = get_value(self._json_config, 100, "segmenter", "params", "max_num_bins")
        self.train_col = self.get_train_col()
        self.train_name = get_value(self._json_config, "train_name", "segmenter", "params", "train_name")
        self.test_col = get_value(self._json_config, self.train_col, "segmenter", "params", "test_col")
        self.test_name = get_value(self._json_config, "test_name", "segmenter", "params", "test_name")
        self.duplicates = self.get_duplicates()
        self.seg_list = get_value(self._json_config, None, "segmenter", "seg_list")

    def get_type(self):
        try:
            seg_type = self._json_config["segmenter"]["type"]
            if seg_type.lower().startswith('s'):
                seg_type = 'static'
            elif seg_type.lower().startswith('d'):
                seg_type = 'dynamic'
            elif seg_type.lower().startswith('v'):
                seg_type = 'value'
            else:
                logging.error("Segmenter type '%s' is not understood. Supported segmenters are %s", seg_type,
                              SUPPORTED_SEGMENTERS)
                sys.exit(1)
            return seg_type
        except KeyError:
            return None

    def get_train_col(self):
        try:
            train_col = self._json_config["segmenter"]["params"]["train_col"]
            # make sure the cat_cols and the train_col do not have an intersection because we process the data first
            # and the train_col won't exist if it gets encoded
            if self._process.dummies_activate:
                validate_intersection_cols(cols1=self._process.dummies_cols,
                                           cols2=[train_col],
                                           cols1_name='cat_cols',
                                           cols2_name='train_col')
            # if train_col is a categorical column, which will be label_encoded, we'd need to backup the column because
            # the data gets to segmenter after it's being processed, i.e. the encoded values will reach to segmenter.
            # Because we only load segment column for making predictions before processing them, we can't use encoded
            # list for loading test data
            if self.type == "value" and self._process.label_encoding_activate and train_col in self._process.label_encoding_cols:
                train_col_name = f"__copy_{train_col}"
                # add train_col to the columns that should be copied
                self._column.copy_cols[train_col] = train_col_name
                # add the column to the list of valid columns
                self._column.valid_cols.append(train_col_name)
                # now add the copied column to the list of column that should be dropped before training
                self._column.drop_from_train.append(train_col_name)
                train_col = train_col_name
            return train_col

        except KeyError:
            return None

    def get_duplicates(self):
        try:
            duplicates = self._json_config["segmenter"]["params"]["duplicates"]
            validate_subset_list(parent_list=SUPPORTED_DUPLICATES_FOR_SEGMENTERS,
                                 parent_name="supported duplicates for segmenters",
                                 subset_list=[duplicates],
                                 subset_name="defined duplicates")
            return duplicates
        except KeyError:
            return "merge"



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import sys
import time
from sklearn.model_selection import GridSearchCV
from sklearn import model_selection
from sklearn.externals.joblib import Parallel


# the code adapted from MNE-Python
class ProgressBar:
    """Generate a command-line progressbar"""
    template = '\r[{0}{1}] {2:.0f}% | {3:.02f} sec | {4} '

    def __init__(self, title='', max_value=1, initial_value=0, max_chars=80, progress_character='#', verbose_bool=True):
        """Instantiate a progressbar object using the default or provided parameters

        :param title: message to include at end of progress bar
        :param max_value: maximum value of process (e.g. number of samples to process, bytes to download, etc.)
        :param initial_value: initial value of process, useful when resuming process from a specific value
        :param max_chars: number of characters to use for progress bar
        :param progress_character: Character in the progress bar that indicates the portion completed
        """
        self.cur_value = initial_value
        self.max_value = float(max_value)
        self.title = title
        self.max_chars = max_chars
        self.progress_character = progress_character
        self._do_print = verbose_bool
        self.start = time.time()

        self.closed = False
        self.update()

    def update(self):
        """Update progressbar with current value of process

        :return: None
        """

        progress = min(float(self.cur_value) / self.max_value, 1.)
        num_chars = int(progress * self.max_chars)
        num_left = self.max_chars - num_chars

        duration = time.time() - self.start

        # the \r tells the cursor to return to the beginning of the line rather than starting a new line.  This allows
        # having a progressbar-style display in the console window.
        bar = self.template.format(self.progress_character * num_chars, ' ' * num_left, progress * 100,
                                   duration, self.title)
        # force a flush because sometimes when using bash scripts and pipes, the output is not printed until after the
        # program exits.
        if self._do_print:
            sys.stdout.write(bar)
            sys.stdout.flush()

        if progress == 1:
            self.close()

        self.cur_value += 1

    def close(self):
        """Finish the progressbar using a linebreak and flushing the output"""
        if not self.closed:
            sys.stdout.write('\n')
            sys.stdout.flush()
            self.closed = True

    def __call__(self, sequence):
        sequence = iter(sequence)
        while True:
            try:
                yield next(sequence)
                self.update()
            except StopIteration:
                return


class GridSearchCVProgressBar(GridSearchCV):
    """Monkey patch Parallel to have a progress bar during grid search"""

    def _get_param_iterator(self):
        """Return ParameterGrid instance for the given param_grid

        :return: ParameterGrid iterator
        """
        iterator = super(GridSearchCVProgressBar, self)._get_param_iterator()
        iterator = list(iterator)
        n_candidates = len(iterator)

        cv = model_selection._split.check_cv(self.cv, None)
        n_splits = getattr(cv, 'n_splits', 3)
        max_value = n_candidates * n_splits

        class ParallelProgressBar(Parallel):
            def __call__(self, iterable):
                bar = ProgressBar(max_value=max_value, title='GridSearch')
                iterable = bar(iterable)
                return super(ParallelProgressBar, self).__call__(iterable)

        # monkey patch
        model_selection._search.Parallel = ParallelProgressBar

        return iterator



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""Given a similarity matrix, and predictions of a classifier, this module extends the prediction to a certain number
defined by the user"""

from operator import itemgetter
import warnings
import logging
import pandas as pd
import numpy as np


class Extenders:

    def __init__(self, mdl, conf):
        """This classes receives a MultilabelBinarizer object, a trained classifier, and a similarity matrix to
        extend the predictions of the classifier."""

        self._mdl = mdl
        self._conf = conf

    @property
    def mdl(self):
        return self._mdl

    @property
    def conf(self):
        return self._conf

    def extend(self, column_name):
        """The main function for extending the predictions.

        :return: An updated Learner classifier, which holds the extended predictions
        """
        logging.info("Extending the predictions")

        # use the previously trained mlb object to binarize the predictions of the model
        prediction_mlb = pd.DataFrame(self._mdl["mlb_obj"].transform(self._mdl["prediction"]), columns=self._mdl["mlb_obj"].classes_)
        # use the similarity matrix to compute the similarity scores for all items in the model and each row in the
        # prediction array
        sim_scores = self.get_similarity_scores(prediction_mlb, self._mdl["sim_matrix"], columns=self._mdl["mlb_obj"].classes_)

        # check whether the number of basket extension in config is larger than the predicted number of baskets
        if self._conf.recommender.num_recs > sim_scores.shape[1]:
            warnings.warn("The number of recommendations provided in the json config exceeds the maximum number of "
                          "recommendations provided by Learner, which is equals to {0}".format(sim_scores.shape[1]), Warning)

        # use the columns of the mlb object, the similarity scores, and the binarized rows to make a tuple
        # note: the sorted_sim_scores does not contain the sorted value yet. It gets sorted in the next step
        sorted_sim_scores = self.tuplize_scores_predictions_for_sorting(prediction_mlb, sim_scores)

        # get the extended predictions
        self._mdl["extended_prediction"] = self.get_extended_prediction(sorted_sim_scores,
                                                                        self._mdl["prediction"],
                                                                        self._conf.recommender.num_recs,
                                                                        column_name=column_name)

        logging.info("Successfully extended the predictions")
        return self._mdl

    @staticmethod
    def get_similarity_scores(mlb_data, similarity_matrix, columns=None):
        """Do a matrix multiplication to get the similarity score for each prediction.

        :param mlb_data: dataframe that contains the binarized predictions
        :param similarity_matrix: the similarity matrix
        :param columns: the columns to use for creating the dataframe
        :return: a pandas dataframe that

        Note: the number of columns of the mlb_data should be equal to the number of rows of the similarity matrix to
        perform the matrix multiplication
        """
        return pd.DataFrame(mlb_data.values.dot(similarity_matrix.values), columns=columns)

    @staticmethod
    def tuplize_scores_predictions_for_sorting(mlb_data, similarity_scores):
        """Use the columns of the mlb object, the similarity scores, and the binarized rows to make a tuple

        :param mlb_data: the pandas dataframe containing binarized predictions
        :param similarity_scores: the pandas dataframe containing similarity scores
        :return: a pandas dataframe in which each element contains a tupe of column name (service code for example),
                 similarity score, and binarized data
        """
        sorted_sim_scores = pd.DataFrame()
        for col in mlb_data.columns:
            col_name = np.full(mlb_data.shape[0], col)
            sorted_sim_scores[col] = list(zip(col_name, similarity_scores[col], mlb_data[col]))
        return sorted_sim_scores

    def get_extended_prediction(self, tupled_similarity_scores, predictions, num_recs, column_name):
        """Sort the predictions and extend them based on the similarity scores

        :param tupled_similarity_scores: pandas dataframe containing the tupled similarity score
        :param predictions: the tupled predictions (the classification results)
        :param num_recs: the target number of recommendations set by the user
        :return: the extended predictions in pandas series format

        Note: to speed up the calculations, extension should only be done on unique baskets
        """
        # a temporary dictionary to hold the rows that are extended and their corresponding extension
        # this helps speed up the extension process by not sorting the predictions that have already been extended
        temp_dict = {}
        extended_prediction = []
        temp_df = pd.DataFrame()

        for i in range(tupled_similarity_scores.shape[0]):
            if predictions[i] in temp_dict:
                extended_prediction.append(temp_dict[predictions[i]])
            else:
                # first sort the recommendation from the classification model, break the tie on similarity score
                row = sorted(tupled_similarity_scores.iloc[i], key=lambda x: x[::-1], reverse=True)

                # now sort the rest according to the similarity score
                row[len(predictions[i]):] = sorted(row[len(predictions[i]):], key=itemgetter(1), reverse=True)

                temp_df = temp_df.append(row)

                # extract the service codes from the tuples
                temp_ext_pred = []
                for item in row[:num_recs]:
                    temp_ext_pred.append(item[0])

                # add the extended predictions in tuple
                extended_prediction.append(tuple(temp_ext_pred))

                # add the prediction and the extension to the temp_dict to avoid redoing things
                temp_dict[predictions[i]] = tuple(temp_ext_pred)

        return pd.Series(extended_prediction, name=column_name)




# filename and full path for saving the prediction files (without segmentation)
PRED_PATH = "{path}{output_name}{dtype_sep}{tag}{sep_timetag}.csv"
PRED_PATH_CHUNK = "{path}{output_name}{dtype_sep}{tag}{sep_timetag}_{index}.csv"
PRED_FILENAME = "{output_name}{dtype_sep}{tag}{sep_timetag}.csv"

# filename and full path for saving the prediction files (with segmentation)
PRED_PATH_SEG = "{path}{output_name}{dtype_sep}{tag}{sep_timetag}_{seg_id}.csv"
PRED_PATH_CHUNK_SEG = "{path}{output_name}{dtype_sep}{tag}{sep_timetag}_{seg_id}{sep_index}.csv"
PRED_FILENAME_SEG = "{output_name}{dtype_sep}{tag}{sep_timetag}_{index}.csv"
PRED_FILENAME_CHUNK_SEG = "{output_name}{dtype_sep}{tag}{sep_timetag}_{seg_id}{sep_index}.csv"


SAVE_MODEL_PATH = "{path}{output_name}_{tag}{sep_timetag}{ext}"
SAVE_MODEL_PATH_SEG = "{path}{output_name}_{tag}_{seg_id}{sep_timetag}{ext}"
SAVE_MODEL_PATH_EPOCH = "{path}{output_name}_{tag}_{epoch}{sep_timetag}{ext}"
LOAD_MODEL_PATH_SEG = "{path}{prefix}_{seg_id}{sep_timetag}{ext}"

FEATURE_IMPORTANCE_PATH = "{output_name}_{tag}_feature_importance{sep_timetag}"
FEATURE_IMPORTANCE_PATH_SEG = "{output_name}_{tag}_{seg_id}_feature_importance{sep_timetag}"

PREDICTIONS_VS_ACTUALS_PLOT = "{path}{output_name}_{tag}_predictions_vs_actuals{sep_timetag}.png"

SHAP_PATH = "{output_name}_{tag}_shap_values{class_name}{sep_timetag}.csv"
SHAP_PLOT = "{path}{output_name}_{tag}_shap_{type}{sep_timetag}.png"
SHAP_PLOT_ClASS = "{path}{output_name}_{tag}_shap_{type}_{class_name}{sep_timetag}.png"

TRAIN_VALIDATION_SPLIT = "{output_name}_{dtype}_split{sep_timetag}.{format}"

LEARNING_RATE = "{path}{output_name}_{tag}_learning_rate{sep_timetag}.png"



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module for constructing the loss functions. Currently, this module implements the LossBuilder class, which
creates the loss objects."""
import sys
from copy import deepcopy

import logging
from torch.nn import CrossEntropyLoss, NLLLoss, MSELoss, L1Loss

from learner.data_worker.data_processor import delete_keys_from_dict


class LossBuilder:
    """Use a dictionary of parameters to build the loss objects. The main method here is build_loss. This method uses
    functional programming to build the loss objects."""
    def __init__(self, loss_dict):
        """Initialize a LossBuilder object using a dictionary of parameters. This dictionary would contain the type of
        the loss function as well as other parameters.

        :param loss_dict: a dictionary that contains all the parameters for creating the loss function.
        """
        self._loss_type = loss_dict["type"]
        self._loss_dict = deepcopy(loss_dict)

    @property
    def loss_type(self):
        return self._loss_type

    @property
    def loss_dict(self):
        return self._loss_dict

    def build_loss(self):
        """The main method to get the loss object using functional programming. Here, we use the loss type to
        get the loss object.

        :return: a loss instance
        """
        delete_keys_from_dict(self.loss_dict, ["type"])
        try:
            return getattr(sys.modules[__name__], self.loss_type)(**self.loss_dict)
        except Exception as e:
            logging.critical(f"Unable to build the loss. The error is: {e}")
            sys.exit(1)




# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This modules implements multiple classes to handle operations related to image data. This module communicates with
other data worker modules to load, validate, and process the image data. Currently, the image_classifier engine uses
the classes in this module"""

import logging

from learner.validator.input_validator import remove_subset_list
from torch.utils.data import DataLoader

from learner.data_worker.image_data_loader import ImageDataLoader
from learner.data_worker.data_set import LearnerTrainImageDataset, LearnerTestImageDataset
from learner.configuration.configuration import Configuration
from learner.data_worker.image_processor import ImageTransformer
from learner.data_worker.data_sampler import DataSampler


class TrainDataManager:
    """This class manages image the training data. The main method in this class is `get_data_loader` method. Unlike
    the class in data_manager module, we need to explicitly call the method to get the data_loader."""
    def __init__(self, conf: Configuration, mdl):
        """Initialize a TrainDataManager object using a conf object and an item in models_dict.

        :param conf: a conf object
        :param mdl: an item of models_dict. We use this to save the model classes.
        """
        self._conf = conf
        self._mdl = mdl

    @property
    def conf(self):
        return self._conf

    @property
    def mdl(self):
        return self._mdl

    def get_df(self):
        """Use the ImageDataLoader class to construct a dataframe from the image data.

        :return: a pandas dataframe that contains information about the path to the images and their corresponding label
        """
        loader = ImageDataLoader(self.conf, self.mdl)
        return loader.get_train_df()

    def get_transform(self):
        """Use the ImageTransformer class to build a Compose object for transforming the images.

        :return: a torchvision Compose object ready to be passed to a Dataset class
        """
        transform = ImageTransformer(self.conf, data_type="train").build_transform()
        return transform

    def get_dataset(self, df, transform):
        """Use LearnerTrainImageDataset class, the dataframe, and the transformer object to obtain a dataset object.

        :params: a pandas dataframe that contains information about the path to the images and their corresponding label
        :param: a torchvision Compose object ready to be passed to a Dataset class
        :return: a LearnerTrainImageDataset object
        """
        dataset = LearnerTrainImageDataset(df, path_col=self.conf.column.path_col, target_col=self.conf.column.target_col, transform=transform)
        return dataset

    def sample(self, df):
        """Since we build a pandas dataframe from all image data, we can also sample that dataframe, if needed, to
        create a validation dataset or reduce the size of the training data.

        :params: a pandas dataframe that contains information about the path to the images and their corresponding label
        :return: the sampled training dataframe
        """
        sample = DataSampler(self.conf, df)
        return sample.sample_data()

    def get_data_loader(self):
        """The main method for getting the train data loader. This method calls other methods in this class to obtain
        all the necessary information for creating a DataLoader object. This includes getting the dataframe, getting
        the transforms, and getting the dataset.

        :return: a PyTorch DataLoader object
        """
        logging.info("Getting the train data loader...")
        df = self.get_df()
        df = self.sample(df)
        transform = self.get_transform()
        dataset = self.get_dataset(df, transform)

        batch_size = self.mdl["params"]["batch_size"] if "batch_size" in self.mdl["params"] and self.mdl["params"]["batch_size"] else self.conf.data.train_batch_size
        return DataLoader(dataset, batch_size=batch_size, shuffle=True)


class ValidationDataManager:
    """This class manages image the validation data. The main method in this class is `get_data_loader` method. Unlike
    the class in data_manager module, we need to explicitly call the method to get the data_loader. The train and
    validation data have some similarities and some differences. For example, the validation data is accompanied by
    labels similar to training data but transformations are similar to the test data."""
    def __init__(self, conf, mdl):
        """Initialize a ValidationDataManager object using a conf object and an item in models_dict.

        :param conf: a conf object
        :param mdl: an item of models_dict. We use this to save the model classes.
        """
        self._conf = conf
        self._mdl = mdl

    @property
    def conf(self):
        return self._conf

    @property
    def mdl(self):
        return self._mdl

    def get_df(self):
        """Use the ImageDataLoader class to construct a dataframe from the image data.

        :return: a pandas dataframe that contains information about the path to the images and their corresponding label
        """
        loader = ImageDataLoader(self.conf, self.mdl)
        return loader.get_validation_df()

    def get_transform(self):
        """Use the ImageTransformer class to build a Compose object for transforming the images.

        :return: a torchvision Compose object ready to be passed to a Dataset class
        """
        transform = ImageTransformer(self.conf, data_type="validation").build_transform()
        return transform

    def get_dataset(self, df, transform):
        """Use LearnerTrainImageDataset class, the dataframe, and the transformer object to obtain a dataset object.
        Please note that we don't have a LearnerValidationImageDataset class because of the similarities between train
        and validation data. However, the transforms passed to the LearnerTrainImageDataset are built for validation
        data, which is different from the one for the training data.

        :params: a pandas dataframe that contains information about the path to the images and their corresponding label
        :param: a torchvision Compose object ready to be passed to a Dataset class
        :return: a LearnerTrainImageDataset object
        """
        dataset = LearnerTrainImageDataset(df, path_col=self.conf.column.path_col, target_col=self.conf.column.target_col, transform=transform)
        return dataset

    def get_data_loader(self):
        """The main method for getting the validation data loader. This method calls other methods in this class to
        obtain all the necessary information for creating a DataLoader object. This includes getting the dataframe,
        getting the transforms, and getting the dataset.

        :return: a PyTorch DataLoader object
        """
        logging.info("Getting the validation data loader...")
        df = self.get_df()
        transform = self.get_transform()
        dataset = self.get_dataset(df, transform)

        return DataLoader(dataset, batch_size=self.conf.data.validation_batch_size, shuffle=False)


class TestDataManager:
    """This class manages image the test data. The main method in this class is `get_data_loader` method. The test and
    validation data have some similarities and some differences. For example, the validation data is accompanied by
    labels unlike test data but transformations are similar."""
    def __init__(self, conf, mdl):
        """Initialize a TestDataManager object using a conf object and an item in models_dict.

        :param conf: a conf object
        :param mdl: an item of models_dict. We use this to save the model classes.
        """
        self._conf = conf
        self._mdl = mdl

    @property
    def conf(self):
        return self._conf

    @property
    def mdl(self):
        return self._mdl

    def get_df(self):
        """Use the ImageDataLoader class to construct a dataframe from the image data.

        :return: a pandas dataframe that contains information about the path to the images as well as some other columns
        """
        loader = ImageDataLoader(self.conf, self.mdl)
        return loader.get_test_df()

    def get_transform(self):
        """Use the ImageTransformer class to build a Compose object for transforming the images.

        :return: a torchvision Compose object ready to be passed to a Dataset class
        """
        transform = ImageTransformer(self.conf, data_type="test").build_transform()
        return transform

    def get_dataset(self, df, transform):
        """Use LearnerTestImageDataset class, the dataframe, and the transformer object to obtain a dataset object. The
        LearnerTestImageDataset behaves differently from the LearnerTrainImageDataset because there's no label in
        test data. Additionally, the LearnerTestImageDataset may need to return additional columns such as id columns
        from the data, which is different from train or validation data.

        :params: a pandas dataframe that contains information about the path to the images as well as some other columns
        :param: a torchvision Compose object ready to be passed to a Dataset class
        :return: a LearnerTestImageDataset object
        """
        cols = remove_subset_list(self.conf.column.valid_cols, [self.conf.column.target_col])
        dataset = LearnerTestImageDataset(df, path_col=self.conf.column.path_col, use_cols=cols, transform=transform)
        return dataset

    def get_data_loader(self):
        """The main method for getting the test data loader. This method calls other methods in this class to
        obtain all the necessary information for creating a DataLoader object. This includes getting the dataframe,
        getting the transforms, and getting the dataset.

        :return: a PyTorch DataLoader object
        """
        logging.info("Getting the test data loader...")
        df = self.get_df()
        transform = self.get_transform()
        dataset = self.get_dataset(df, transform)

        return DataLoader(dataset, batch_size=self.conf.data.test_batch_size, shuffle=False)



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements the DeepRegressor module. Please note that the majority of the functionality is implemented
in the base class"""

import sys
import logging

from learner.configuration.configuration import Configuration
from learner.utilities.timer import timeit
from learner.model_manager.scoring_manager import RegressorScorer
from learner.combine.combining_manager import MeanRegressorCombiner
from learner.engines.base_deep_engine import BaseDeep
from learner.combine.combining_manager import TriadCombiner


class DeepRegressor(BaseDeep):
    def __init__(self, conf: Configuration):
        """Take a conf object (from the Configuration class) to run a deep_regressor engine.

        :param conf: a conf object
        """
        super().__init__(conf)

    @timeit("compute the predictions scores")
    def score_prediction(self):
        """Score the predictions of the models.

        :return: None
        """
        if self._conf.data.validation_score_activate:
            logging.info("Computing the prediction scores...")
            scorer = RegressorScorer(self.conf, self.models_dict)
            scorer.score()





# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module for constructing the models_dict and loading models/objects"""

import sys
import re
import glob
import os
from datetime import datetime
import logging
import pickle


class ModelInitializer:
    """Construct the models_dict."""

    def __init__(self, conf):
        """Obtain a conf object and make a call to build_models_dict to construct models_dict."""

        self._conf = conf
        # this is silly but in the past we needed to modify the models_dict and we had a method called build_models_dict
        # I'm keeping things this way, in case we may need to implement a similar method
        self.models_dict = self._conf.model.models_dict

    @property
    def conf(self):
        return self._conf


class ModelLoader:
    """Load pickled objects based on path to .sav file"""

    def __init__(self, conf):
        """Initiate ModelLoader using a conf object"""

        self._conf = conf

    @property
    def conf(self):
        return self._conf

    def load_model(self, path_to_pickle):
        """Load a pickled object based on a file path. Learner uses the data and combine sections in the new
        configuration file after loading the pickled configuration.

        :param path_to_pickle: the path to the pickle object
        :return: a previously pickled configuration object, models_dict, processor object and sep_timetag
        """
        try:
            with open(path_to_pickle, "rb") as pickled_file:
                conf, models_dict, processor, feature_engineering = pickle.load(pickled_file)
                # we want to use the current data parameters
                conf.data = self._conf.data
                # we want to use the current combine parameters
                conf.combine = self._conf.combine
                # we want to use the current workspace parameters
                conf.workspace = self._conf.workspace
                # we load the model, we've made 0 predictions
                conf.model.nrows_score = 0
                conf.timetag = self._conf.timetag
                sep_timetag = conf.sep_timetag
                conf.sep_timetag = self._conf.sep_timetag
                # we want to use the current connection object
                conf.connection = self._conf.connection
                logging.info("Successfully loaded pickled object")
                return conf, models_dict, processor, feature_engineering, sep_timetag
        except FileNotFoundError:
            logging.exception("Output path or name doesn't exist. Check json configuration.")
            sys.exit(1)

    def load_combiner(self, path_to_pickle):
        """Load a pickled combiner object. The combine object has a nested conf object in it. When loading the combiner
        object we replace the old conf object with the current conf object.

        :param path_to_pickle: the path to the pickle object
        :return: the loaded combiner object
        """
        with open(path_to_pickle, "rb") as pickled_file:
            combiner = pickle.load(pickled_file)
            combiner.conf = self._conf
            return combiner


def get_pickle_dir(path_to_pickle, ext="sav"):
    """Load pickled object and retrieve its directory and model prefix.
    If path_to_pickle includes a directory with a time tag, a .sav object will be retrieved that corresponds to
    that time tag, directory path to this object, and model prefix
    If path_to_pickle is just a path that doesn't include time tag, the directory to the latest pickled object
    will be retrieved.

    :param path_to_pickle: directory path to an engine output or the directory path to pickled object.
    :param ext: the file extensions to use to find the most recent models, e.g. sav, pkl, etc. Do not include "." here.
    :return: directory path to pickle, pickled object, prefix. The outputs are all lists
    """
    # retrieve pickled object that corresponds to the latest time tag
    pickled_files = []
    # save the list of prefixes
    prefixes = []
    # a list containing the directory path to pickle objects
    dir_paths_to_pickle = []

    try:
        # Check if a path has any .sav files
        if glob.glob(path_to_pickle+f'*.{ext}'):
            pickled_files.append([file for file in glob.iglob(path_to_pickle+f"*.{ext}")][-1])
            # if the folder has a timetag, the prefix will be obtained differently
            timetag = re.search(r'[0-9]{4}\-[0-9]{2}\-[0-9]{2}\-[0-9]{2}\-[0-9]{2}', path_to_pickle)
            if timetag:
                prefixes.append("_".join(os.path.basename(os.path.dirname(path_to_pickle)).split("_")[:-1]))
            else:
                prefixes.append(os.path.basename(os.path.dirname(path_to_pickle)))
            return [path_to_pickle], pickled_files, prefixes

        # Retrieve timetags from all existing directories
        timetags = []
        for folder in os.listdir(path_to_pickle):
            timetag = re.search(r'[0-9]{4}\-[0-9]{2}\-[0-9]{2}\-[0-9]{2}\-[0-9]{2}', folder)
            if timetag:
                timetags.append(timetag.group())

        # sort directories by time tag from latest to earliest
        sorted_timetags = sorted(timetags, key=lambda x: datetime.strptime(x, '%Y-%m-%d-%H-%M'),
                                 reverse=False)
        # retrieve the directory with the most recent timetag
        latest_pickled_folder = [folder_name for folder_name in os.listdir(path_to_pickle) if
                                 sorted_timetags[-1] in folder_name]

        for folder in latest_pickled_folder:
            for file in glob.iglob(path_to_pickle+folder + "/" + f"*.{ext}"):
                if file:
                    pickled_files.append(file)
                    prefixes.append("_".join(os.path.basename(os.path.basename(folder)).split("_")[:-1]))
                    dir_paths_to_pickle.append(path_to_pickle + folder + "/")
                    break
        return dir_paths_to_pickle,  pickled_files, prefixes
    except IndexError:
        # if we did not create triad but looking for it, we will get here. In that case, we don't want to raise an
        # error
        if ext == "triad":
            return [path_to_pickle], pickled_files, prefixes
        else:
            logging.error("Unable to find the folder where the saved models are stored. Please update your "
                          "configuration file and try again. Exiting...")
            sys.exit(1)



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module includes Learner's deep regressors. The pattern for deep regressors is no different than the standard
regressirs. Here, we do not directly wrap sklearn models, etc.
"""

import sys
import abc
import logging

import torch
import torch.nn as nn

from learner.model_manager.layer_manager import LayerBuilder
from learner.model_manager.loss_manager import LossBuilder
from learner.model_manager.optimizer_manager import OptimizerBuilder
from learner.model_manager.scheduler_manager import SchedulerBuilder, LRFinderScheduler
from learner.model_manager.deep_loop_manager import DeepLoopManager
from learner.callback_manager.callback import MetricLogger, ModelSaver, LRFinder
from learner.data_worker.data_mover import move_to_device


class AbstractDeepRegressor:
    """The AbstractDeepRegressor class. This class does not implement any functionality, it defines the structure of
    the child classes.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):  # pragma: no cover
        self.mdl = None

    @staticmethod
    def learner_fit(tag, mdl, conf, train_loader, validation_loader):  # pragma: no cover
        pass

    @staticmethod
    def learner_predict(X, mdl, conf, carry_data=True, full_data=None, cols_to_carry=None):  # pragma: no cover
        pass

    @staticmethod
    def learner_find_learning_rate(tag, mdl, conf, train_loader):
        pass


class DeepRegressorHandler(AbstractDeepRegressor):
    """The handlers class calls the appropriate method of the Regressor classes for the type defined in mdl.
    """
    @abc.abstractmethod
    def __init__(self):
        super(AbstractDeepRegressor, self).__init__()  # pragma: no cover

    @staticmethod
    def learner_fit(tag, mdl, conf, train_loader, validation_loader):
        return getattr(sys.modules[__name__], mdl["type"]).learner_fit(tag, mdl, conf, train_loader, validation_loader)

    @staticmethod
    def learner_predict(X, mdl, device, carry_data=True, full_data=None, cols_to_carry=None):
        return getattr(sys.modules[__name__], mdl["type"]).learner_predict(X, mdl, device, carry_data, full_data, cols_to_carry)

    @staticmethod
    def learner_find_learning_rate(tag, mdl, conf, train_loader):
        return getattr(sys.modules[__name__], mdl["type"]).learner_find_learning_rate(tag, mdl, conf, train_loader)


class DeepRegressor:
    """The DeepRegressor class to build a deep learning regression model for tabular data.
    """
    @staticmethod
    def learner_fit(tag, mdl, conf, train_loader, validation_loader):
        """First, instantiate the LearnerDeepRegressor model. Then build the loss function, the optimizer, and
        the scheduler. After that, use DeepTrainManager class to train and validate the model. In the end, return the
        trained model.

        :param tag: the model tag, this is an arbitrary string defined by the user.
        :param mdl: an item of the models_dict.
        :param conf: a conf object
        :param train_loader: the train_loader to loop through the training data.
        :param validation_loader: the validation_loader to loop through the validation data.
        :return: the trained model.
        """
        model = LearnerDeepRegressor(conf=conf, mdl=mdl)
        criterion = LossBuilder(mdl["params"]["loss"]).build_loss()
        optimizer = OptimizerBuilder(list(model.parameters()), mdl["params"]["optimizer"]).build_optimizer()
        scheduler = SchedulerBuilder(mdl["params"]["scheduler"] if "scheduler" in mdl["params"] else None).build_scheduler(optimizer)

        train_manager = DeepLoopManager(conf,
                                        model,
                                        criterion,
                                        optimizer,
                                        scheduler,
                                        mdl["params"]["epochs"],
                                        tag=tag,
                                        mdl=mdl,
                                        train_loader=train_loader,
                                        validation_loader=validation_loader,
                                        callbacks=[MetricLogger(), ModelSaver()])
        train_manager.train()
        return model

    @staticmethod
    def learner_find_learning_rate(tag, mdl, conf, train_loader):
        """First, instantiate the LearnerDeepRegressor model. Then build the loss function and the optimizer.
        Set the initial learning rate of the optimizer using the parameter defined in learning_rate_params of the
        configuration file. Also, compute the number of epochs using the number of iterations needed and the number of
        batches in the train_loader. After that, use DeepTrainManager class to find the learning rate. Here, we pass
        the LRFinder callback to the DeepTrainManager class. This callback handles things related to the learning rate
        parameters such as stopping the training loop, etc. In the end, return the trained model.

        :param tag: the model tag, this is an arbitrary string defined by the user.
        :param mdl: an item of the models_dict.
        :param conf: a conf object
        :param train_loader: the train_loader to loop through the training data.
        :return: the trained model.
        """
        model = LearnerDeepRegressor(conf=conf, mdl=mdl)

        criterion = LossBuilder(mdl["params"]["loss"]).build_loss()
        # set the initial lr to be start_lr irrespective of what's being set
        mdl["params"]["optimizer"]["lr"] = conf.model.lr_start_lr
        # determine how many epochs we need depending on the size of data and num_iter
        mdl["params"]["epochs"] = int(conf.model.lr_num_it / len(train_loader)) + 1

        optimizer = OptimizerBuilder(list(model.parameters()),
                                     mdl["params"]["optimizer"]).build_optimizer()

        scheduler = LRFinderScheduler(optimizer=optimizer,
                                      start_lr=conf.model.lr_start_lr,
                                      end_lr=conf.model.lr_end_lr,
                                      num_it=conf.model.lr_num_it)

        train_manager = DeepLoopManager(conf,
                                        model,
                                        criterion,
                                        optimizer,
                                        scheduler,
                                        mdl["params"]["epochs"],
                                        tag=tag,
                                        mdl=mdl,
                                        train_loader=train_loader,
                                        callbacks=[MetricLogger(),
                                                    LRFinder(conf=conf,
                                                             scheduler=scheduler,
                                                             stop_div=conf.model.lr_stop_div,
                                                             tag=tag,
                                                             mdl=mdl)])
        train_manager.train()
        return model

    @staticmethod
    def learner_predict(X, mdl, device, carry_data=True, full_data=None, cols_to_carry=None):
        """Use the input data and a mdl dictionary along with other arguments to get the predictions. The value
        for the "model" key im mdl dictionary contains the trained model. This methods accepts the "device" argument
        as well. We move the data and the model to the appropriate device before making the predictions. This would
        enable making predictions using CPU or GPU.

        :param X: the input data for making the predictions.
        :param mdl: an item of the models_dict. This dictionary should contain the trained model.
        :param device: the device, this could be "cuda" or "cpu". The values are validated before they reach here.
        :param carry_data: a flag to indicate whether we should carry additional data or not.
        :param full_data: a pandas dataframe with additional data to send back if requested.
        :param cols_to_carry: The list of columns to select from the full dataset and send back if requested.
        :return: the predictions classes in numpy array as well as the selected columns from the data or None depending on the input.
        """
        logging.info("Making predictions using the %s model...", mdl["type"])
        with torch.no_grad():
            X = move_to_device(X, device)
            # put the model to evaluation mode. This is very important otherwise the results will not be correct
            mdl["model"].to(device).eval()
            pred = mdl["model"](*X).cpu()
            if carry_data:
                if full_data is not None and cols_to_carry:
                    return pred, full_data[cols_to_carry]
            return pred, None


class LearnerDeepRegressor(nn.Module):
    """The class to implement a deep regressor model from scratch using the parameters defined by the user. While the
    LearnerDeepRegressor model is quite flexible, it has a specific architecture. It passes all the categorical
    features through embedding layers and the continuous features through a batchnorm layer. It then concatenates the
    outputs and pass them through a series of fully connected layers. The user can/should define the size of the
    embedding layers and all the details about the fully connected layers."""

    def __init__(self, conf, mdl):
        """Initialize a LearnerDeepRegressor object using the input arguments.

        :param conf: a conf object.
        :param mdl: an item of models_dict.
        """
        super(LearnerDeepRegressor, self).__init__()
        self._conf = conf
        self._mdl = mdl

        self.embeddings = nn.ModuleList([nn.Embedding(num_input, num_features) for _, (num_input, num_features) in self._mdl["params"]["embedding_sizes"].items()])
        self.num_embeddings = sum(e.embedding_dim for e in self.embeddings)

        self.embeddings_dropout = nn.Dropout(self._mdl["params"]["embedding_dropout"])
        self.batchnorm_continuous = nn.BatchNorm1d(self._mdl["n_continuous"])

        layer_builder = LayerBuilder(raw_list=self._mdl["params"]["fully_connected_layers"],
                                     initial_size=self._mdl["n_continuous"] + self.num_embeddings)
        fully_connected_layers = layer_builder.build_layers()
        self.fully_connected_layers = nn.Sequential(*fully_connected_layers)

    def forward(self, x_cat, x_cont):
        """If there are categorical features, we pass them through embedding layers. If there are continuous features,
        we pass them through a single BatchNorm1d layer. Then we concatenate the output and pass it through the
        fully connected layer.

        :param x_cat: a PyTorch tensor containing the categorical features. These should contain the indices ready for embedding
        :param x_cont: a PyTorch tensor containing the continuous features. These should already be processes and normalized
        :return:
        """
        if self.num_embeddings != 0:
            x = [e(x_cat[:, i]) for i, e in enumerate(self.embeddings)]
            x = torch.cat(x, 1)
            x = self.embeddings_dropout(x)
        if self._mdl["n_continuous"] != 0:
            x_cont = self.batchnorm_continuous(x_cont)
            x = torch.cat([x, x_cont], 1) if self.num_embeddings != 0 else x_cont
        x = self.fully_connected_layers(x)
        # TODO: revisit this later since it could have performance impact
        x = x.double()
        x = torch.squeeze(x, 1)
        if self._mdl["params"]["y_range"] is not None:
            x = (self._mdl["params"]["y_range"][1] - self._mdl["params"]["y_range"][0]) * torch.sigmoid(x) + self._mdl["params"]["y_range"][0]
        return x



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements all the dataset classes for tabular, image and potentially other data types. For tabular data,
we have a separate class for train, validation, and test data. For image data, one class handles train and validation
data and another one handles test data"""

from collections import OrderedDict

import torch
from torch.utils.data import Dataset
from torchvision import datasets

from learner.configuration.configuration import Configuration
from learner.data_worker.data_loader import remove_subset_list


class LearnerTrainTabularDataset(Dataset):
    """A dataset class for loading the train tabular data. This class would expect a pandas dataframe.
    """
    def __init__(self, df, conf: Configuration, mdl):
        """Initialize a LearnerTrainTabularDataset object using the input arguments. Here, we call the method
        _get_embedding_sizes to get the embedding sizes for the categorical features. We use the mdl to save all the
        necessary information that we may need later on.

        :param df: a pandas dataframe of the training data.
        :param mdl: an item (value) in models_dict.
        """
        self.df = df
        self.mdl = mdl
        self.conf = conf
        self._get_embedding_sizes()
        # it may not be necessary to use an OrderedDict here but just to be safe
        self.mdl["params"]["embedding_sizes"] = OrderedDict(self.mdl["params"]["embedding_sizes"])
        self.mdl["n_continuous"] = len(remove_subset_list(self.mdl["train_features"], self.mdl["params"]["embedding_sizes"].keys()))
        self._get_classes()

    @staticmethod
    def embedding_size_rule(n_cat):
        """Accept the number of categorical data and return the recommended size of the embedding layer for that
        feature. We've taken this from the fastai library. The user can define their own embedding size for each
        feature.

        :param n_cat: number of categories in the feature
        :return: the computed size for the embedding layer
        """
        return min(600, round(1.6 * n_cat ** 0.56))

    def _get_embedding_sizes(self):
        """First, we check if embedding_sizes is self.mdl["params"] is None or not. If it is (meaning that the user
        did not want to overwrite the default values) we set it to an empty dictionary (please see the in-line comment
        to understand why). We then iterate through each categorical column, to build the embedding_size dictionary.
        In this dictionary, the key is the name of the column. The values are tuples in which the first element is the
        input size and the second element is the embedding size like so: {"col1": (3, 4)}.

        :return: None
        """
        # this is important. For some reason, a default {} in the configuration would populate other mdl - super strange
        if self.mdl["params"]["embedding_sizes"] is None:
            self.mdl["params"]["embedding_sizes"] = {}
        for col in self.conf.process.label_encoding_cols:
            # if the user defines, embedding size for a column, we take it
            if self.mdl["params"]["embedding_sizes"] and col in self.mdl["params"]["embedding_sizes"]:
                # + 1 if for unseen levels
                self.mdl["params"]["embedding_sizes"][col] = (self.df[col].nunique() + 1,
                                                              self.mdl["params"]["embedding_sizes"][col])
            # if not, we compute it
            else:
                self.mdl["params"]["embedding_sizes"][col] = (self.df[col].nunique() + 1,
                                                              self.embedding_size_rule(self.df[col].nunique()))

    def _get_classes(self):
        """In DeepClassifier, we need to get the classes. We will use this information later on.

        :return: None
        """
        if self.conf.engine == "DeepClassifier":
            self.mdl["classes"] = sorted(list(self.df[self.conf.column.target_col].unique()))

    def __len__(self):
        """Get the number of items in the dataframe.

        :return: the number of items in the dataframe
        """
        return self.df.shape[0]

    def __getitem__(self, idx):
        """In deep engines, we need to separate continuous and categorical features because we do embedding for
        categorical features. Here, we get the target, categorical, and continuous features given the index. For the
        features we return a list. This is important to remember. We use the embedding sizes dictionary to find the
        categorical features. We then convert the features into PyTorch tensors. We could potentially convert the
        target into a tensor but that's not needed.

        :param idx: the index of the item, this is the row number.
        :return: a tuple where the first item is a list of categorical and continuous features and the second item is the target.
        """
        # at ensures the dtype remains unchanged
        target = self.df.at[idx, self.conf.column.target_col]
        cat_features = self.df.iloc[idx][self.mdl["params"]["embedding_sizes"].keys()]
        cont_features = self.df.iloc[idx][remove_subset_list(self.mdl["train_features"],
                                                             self.mdl["params"]["embedding_sizes"].keys())]

        return [torch.tensor(cat_features).long(), torch.tensor(cont_features).float()], target


class LearnerValidationTabularDataset(Dataset):
    """A dataset class for loading the validation tabular data. This class would expect a pandas dataframe. Unlike the
    Image dataset classes, we need a separate class for the validation tabular data because in train dataset we get the
    embedding sizes and we use that in the validation and test datasets.
    """
    def __init__(self, df, conf: Configuration, mdl):
        """Initialize a LearnerValidationTabularDataset object using the input arguments. Here, assume that the
        embedding sizes dictionary for the categorical features is already populated. We use the mdl to save all the
        necessary information that we may need later on.

        :param df: a pandas dataframe of the training data.
        :param mdl: an item (value) in models_dict.
        """
        self.df = df
        self.mdl = mdl
        self.conf = conf

    def __len__(self):
        """Get the number of items in the dataframe.

        :return: the number of items in the dataframe
        """
        return self.df.shape[0]

    def __getitem__(self, idx):
        """In deep engines, we need to separate continuous and categorical features because we do embedding for
        categorical features. Here, we get the target, categorical, and continuous features given the index. For the
        features we return a list. This is important to remember. We use the embedding sizes dictionary to find the
        categorical features. We then convert the features into PyTorch tensors. We could potentially convert the
        target into a tensor but that's not needed.

        :param idx: the index of the item, this is the row number.
        :return: a tuple where the first item is a list of categorical and continuous features and the second item is the target.
        """
        # at ensures the dtype remains unchanged
        target = self.df.at[idx, self.conf.column.target_col]
        cat_features = self.df.iloc[idx][self.mdl["params"]["embedding_sizes"].keys()]
        cont_features = self.df.iloc[idx][remove_subset_list(self.mdl["train_features"],
                                                             self.mdl["params"]["embedding_sizes"].keys())]

        return [torch.tensor(cat_features).long(), torch.tensor(cont_features).float()], target


class LearnerTestTabularDataset(Dataset):
    """A dataset class for loading the test tabular data. This class would expect a pandas dataframe. The difference
    between the validation and test dataset is that the test dataset does not expect/load any targets.
    """
    def __init__(self, df, conf: Configuration, mdl):
        """Initialize a LearnerTestImageDataset object using the input arguments.

        :param df: a pandas dataframe that contains the path to the images and their corresponding label.
        :param mdl: an item (value) in models_dict.
        """
        self.df = df
        self.mdl = mdl
        self.conf = conf

    def __len__(self):
        """Get the number of items in the dataframe.

        :return: the number of items in the dataframe
        """
        return self.df.shape[0]

    def __getitem__(self, idx):
        """In deep engines, we need to separate continuous and categorical features because we do embedding for
        categorical features. Here, we get the categorical continuous features given the index. For the
        features we return a list. This is important to remember. We use the embedding sizes dictionary to find the
        categorical features. We then convert the features into PyTorch tensors. Here, unlike train and validation
        datasets, we return all the raw data as well. This data is used later on to return additional data if requested.

        :param idx: the index of the item, this is the row number.
        :return: a tuple where the first item is a list of categorical and continuous features and the second item is the target.
        """
        cols = remove_subset_list(self.conf.column.valid_cols, [self.conf.column.target_col])
        data = self.df[cols].iloc[idx].to_dict()
        cat_features = self.df.iloc[idx][self.mdl["params"]["embedding_sizes"].keys()]
        cont_features = self.df.iloc[idx][remove_subset_list(self.mdl["train_features"],
                                                             self.mdl["params"]["embedding_sizes"].keys())]

        return [torch.tensor(cat_features).long(), torch.tensor(cont_features).float()], data


class LearnerTrainImageDataset(Dataset):
    """A dataset class for loading the train and validation image data. This class uses a pandas dataframe to
    identify and load images along with their corresponding labels.
    """
    def __init__(self, df, path_col, target_col, transform=None):
        """Initialize a LearnerTrainImageDataset object using the input arguments.

        :param df: a pandas dataframe that contains the path to the images and their corresponding label.
        :param path_col: the column that contains the path to the images.
        :param target_col: the columns that contains the label.
        :param transform: a Compose object from pytorch. This is prepared by other classes/module and depend on user input.
        """
        self.df = df
        self.transform = transform
        self.path_col = path_col
        self.target_col = target_col

    def __len__(self):
        """Get the number of items in the dataframe.

        :return: the number of items in the dataframe
        """
        return self.df.shape[0]

    def __getitem__(self, idx):
        """Load a single image from the dataframe given the index. If transform is defined, pass the PIL.Image.Image
        object through the transform. In the end, return the image data and the corresponding label.

        :param idx: the index of the item, this is the row number.
        :return: the image data and the corresponding label.
        """
        path, target = self.df.iloc[idx][[self.path_col, self.target_col]]
        sample = datasets.folder.default_loader(path)
        if self.transform is not None:
            sample = self.transform(sample)
        return sample, target


class LearnerTestImageDataset(Dataset):
    """A dataset class for loading the test image data. This class uses a pandas dataframe to identify and load images.
     This class returns the image data as well as additional data. The additional data is returned in dictionary format
     because of the limitation imposed by PyTorch. That data is converted to dataframe at a later time.
    """
    def __init__(self, df, path_col, use_cols=None, transform=None):
        """Initialize a LearnerTestImageDataset object using the input arguments.

        :param df: a pandas dataframe that contains the path to the images and their corresponding label.
        :param path_col: the column that contains the path to the images.
        :param use_cols: the additional columns that should be returned from the dataframe.
        :param transform: a Compose object from pytorch. This is prepared by other classes/module and depend on user input.
        """
        self.df = df
        self.path_col = path_col
        self.use_cols = use_cols
        self.transform = transform

    def __len__(self):
        """Get the number of items in the dataframe.

        :return: the number of items in the dataframe
        """
        return self.df.shape[0]

    def __getitem__(self, idx):
        """Load a single image from the dataframe given the index. If transform is defined, pass the PIL.Image.Image
        object through the transform. We also get the additional columns from the data. In the end, return the image
        data and the additional data.

        :param idx: the index of the item, this is the row number.
        :return: the image data and the corresponding label.
        """
        path = self.df.iloc[idx][self.path_col]
        data = self.df.iloc[idx][self.use_cols].to_dict()
        sample = datasets.folder.default_loader(path)
        if self.transform is not None:
            sample = self.transform(sample)
        return sample, data



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""Implement functions that help with adding timing information to different methods and functions."""

import time
import functools
import logging


def timeit(message=None):
    def timeit_decorator(method):
        """A decorator function that takes a method/function and computes the the elapsed time

        :param method: a method/function
        :return: the elapsed_time
        """
        @functools.wraps(wrapped=method)
        def elapsed_time(*args, **kwargs):
            nonlocal message
            start_time = time.time()
            result = method(*args, **kwargs)
            end_time = time.time()
            time_diff = end_time - start_time
            # because we only show 2 digits after the floating point, we skip the message if the time taken is less
            # than 0.01 seconds
            if time_diff > 0.01:
                if not message:  # pragma: no cover
                    message = "run " + method.__name__
                logging.info('Timer: it took {0:.2f} seconds to {1}'.format(time_diff, message))
            return result
        return elapsed_time
    return timeit_decorator
